from pystreaming.video.enc import Encoder
from pystreaming.video.dist import Distributor
from pystreaming.video.req import Requester
from pystreaming.video.dec import Decoder
from pystreaming.video.pub import Publisher
from pystreaming.video.sub import Subscriber
from pystreaming.video.handlers import display, collate, dispfps
from pystreaming.video.patterns import Streamer, Worker, Collector
from pystreaming.video.testimages.imageloader import (
    IMAG_S,
    IMAG_M,
    IMAG_L,
    TEST_S,
    TEST_M,
    TEST_L,
    loadimage,
)



class Empty(Exception):
    """Raised when trying to pop from an empty queue"""

    def __init__(self, calltype):
        super().__init__(f"Tried to pop from an empty {calltype}")


class CircularList:
    """Fixed size list with push-overwrite capabilities."""

    def __init__(self, maxsize=10):
        """Initialize a Circular List with maximum size.

        Args:
            maxsize (int, optional): Maximum number of elements the list
                can store before it begins to overwrite old elements.
                Defaults to 10.

        Raises:
            ValueError: Raised if maxsize argument is not a positve integer.
        """
        if maxsize <= 0:
            raise ValueError("Maxsize must be greater than zero")

        self._front = 0  # inclusive
        self._back = 0  # exclusive
        self.size = 0

        self.maxsize = maxsize
        self._array = [None] * maxsize

    def push(self, item):
        """Push an element FIFO (queue) style.

        Args:
            item (pyobj): Object to push.
        """
        if self.size == self.maxsize:  # overwrite old elements
            assert self._front == self._back
            self._array[self._back] = item
            self._back = (self._back + 1) % self.maxsize
            self._front = (self._front + 1) % self.maxsize
        else:
            self._array[self._back] = item
            self._back = (self._back + 1) % self.maxsize
            self.size += 1

    def pop(self):
        """Pop an element off FIFO (queue) style.

        Raises:
            Empty: Raised when there are no elements in the queue.

        Returns:
            pyobj: The first element in the list.
        """
        if self.size == 0:
            raise Empty(self.__class__.__name__)
        else:
            ret = self._array[self._front]
            self.size -= 1
            self._front = (self._front + 1) % self.maxsize
            return ret

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        """Gets the nth element of the list.
        Elements at the front of the list (would be popped off) are at index 0.

        Args:
            idx (int): Index desired to be modified.

        Raises:
            IndexError: Raise if the index is out of range.

        Returns:
            pyobj: The object desired.
        """
        if idx >= self.size or idx < 0:
            raise IndexError(f"list index {idx} out of range: [0, {self.size})")
        return self._array[(self._front + idx) % self.maxsize]

    def __setitem__(self, idx, new_val):
        """Sets the nth element of the list.
        Elements at the front of the list (would be popped off) are at index 0.

        Args:
            idx (int): Index desired to be modified.
            new_val (pyobj): New object to emplace.

        Raises:
            IndexError: Raise if the index is out of range.
        """
        if idx >= self.size or idx < 0:
            raise IndexError(f"list index {idx} out of range: [0, {self.size})")
        self._array[(self._front + idx) % self.maxsize] = new_val

    def __repr__(self):
        return self._array.__repr__() + f" front: {self._front} back: {self._back}"

    def full(self):
        """Check whether the current list size is the maximum size.

        Returns:
            boolean: True if full else False.
        """
        return self.size == self.maxsize



import cv2
import time
from collections import deque
from pystreaming.listlib.circulardict import CircularOrderedDict


def display(handler, BGR=True, getter=None):
    """Yield frames from a generator and display using OpenCV.

    Args:
        handler (generator): Generator that yields data.
        BGR (bool, optional): Set to False if frame is RGB format. Defaults to True.
        getter (function, optional): Returns the frame from the return type of handler.
            Defaults to None, which acts as an identity.
    """
    for data in handler:
        frame = data if getter is None else getter(data)
        if not BGR:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        cv2.imshow("pystreaming-display", frame)
        if cv2.waitKey(1) & 0xFF == 27:  # esc pressed
            break


def collate(handler, buffer=10, getter=None):
    """Reorder frames that are mixed in transit.

    Args:
        handler (generator): Generator that yields data.
        buffer (int, optional): Size of collate buffer. Defaults to 10.
        getter (function, optional): Returns (frame, meta, idx) from the return type of handler.
            Defaults to None, which acts as an identity.

    Yields:
        tuple(np.ndarray, int): (frame, meta, idx)
    """
    initsize = int(buffer * 1.2)
    collate = CircularOrderedDict(initsize)
    for data in handler:
        frame, meta, idx = data if getter is None else getter(data)
        if idx == 0:  # Restart collate if stream restart occurs
            collate = CircularOrderedDict(initsize)
        collate.insert_end(idx, (frame, meta))
        if len(collate) >= buffer:
            most_recent = min(collate.keys())
            yield collate[most_recent] + (most_recent,)
            collate.delete(most_recent)


def dispfps(handler, n=100):
    """Average iterations per second over last n iterations.

    Args:
        handler (generator): Generator that yields data.
        n (int, optional): Number of iterations to average over. Defaults to 100.

    Yields:
        pyobj: Forwards data through handler
    """
    times = deque()
    for data in handler:
        end = time.time()
        times.append(end)
        if len(times) > n:
            diff = end - times.popleft()
            print(f"\rFPS: {(n / diff):.3f}", end="")
        yield data



from collections import OrderedDict


class CircularOrderedDict:
    def __init__(self, maxsize):
        """A dictionary that never exceeds a given size. If the size limit is reached,
        the least recently inserted element is removed.

        Args:
            maxsize (int): Maximum size of the dictionary.

        Raises:
            ValueError: Raised if maxsize argument is not a positve integer.
        """
        if maxsize <= 0:
            raise ValueError("Maxsize must be greater than zero")
        self.dict = OrderedDict()
        self.maxsize = maxsize

    def pop_front(self):
        """Pop key, value pair from front of dictionary.

        Returns:
            tuple: (key, value)
        """
        return self.dict.popitem(last=False)

    def insert_end(self, key, value):
        """Insert key-value pair into end of dictionary. If the key already
        exists, the key-value pair will be moved to the end of the dictionary
        and the value will be updated.

        Args:
            key (pyobj): Key in key-value pair.
            value (pyobj): Value in key-value pair.
        """
        if key in self.dict:
            self.dict.pop(key)
        self.dict[key] = value
        if len(self.dict) > self.maxsize:
            self.dict.popitem(last=False)

    def delete(self, key):
        """Delete an entry from the dictionary

        Args:
            key (pyobj): Key to delete.
        """
        del self.dict[key]

    def keys(self):
        """Retrieve all keys in the dictionary, from earliest to latest.

        Returns:
            odict_keys: Iterable object of keys.
        """
        return self.dict.keys()

    def __len__(self):
        """Number of key-value pairs in the dictionary.

        Returns:
            int: Number of k-v pairs.
        """
        return len(self.dict)

    def __setitem__(self, key, value):
        """Update a key-value pair, without changing its position in the
        dictionary.

        Args:
            key (pyobj): Key in the key-value pair.
            value (pyobj): Value in key-value pair.

        Raises:
            KeyError: Key was not found in the dictionary.
        """
        if key not in self.dict:
            raise KeyError(str(key) + " use insert_end to add an element")
        self.dict[key] = value

    def __getitem__(self, key):
        """Retrieve a value from the dictionary.

        Args:
            key (pyobj): Key in the key-value pair.

        Returns:
            pyobj: Value in the key-value pair, if the key exists.
        """
        return self.dict[key]

    def __repr__(self):
        return self.dict.__repr__()



import zmq
import zmq.asyncio
import asyncio
import multiprocessing as mp
from pystreaming.video import STOPSTREAM, FRAMEMISS, TRACKMISS


"""Stop on STOPSTREAM, or TRACKMISS
Continue on FRAMEMISS
Wait at most TIMEOUT for receiving a response?
    have to use async poll for this behavior
"""


async def aioreq(context, source, track, drain):
    socket = context.socket(zmq.REQ)
    socket.connect(source)
    track_bytes = bytes(track, "utf-8")
    while True:
        # rate limit to 30x a second => 90x requests a sec by default with 3 threads
        await asyncio.sleep(1 / 30)
        await socket.send(track_bytes)
        buf = await socket.recv()
        idx = await socket.recv_pyobj()
        if idx == STOPSTREAM:
            raise StopAsyncIteration(f"Stop stream signal received... Exiting...")
        if idx == FRAMEMISS:
            continue  # throw away if no frame available
        if idx == TRACKMISS:
            raise StopAsyncIteration(
                f'Track "{track}" was not recognized by the Distributor.'
            )
        try:
            await drain.send(buf, copy=False, flags=zmq.SNDMORE | zmq.NOBLOCK)
            await drain.send_pyobj(idx, flags=zmq.NOBLOCK)
        except zmq.error.Again:
            pass  # ignore send misses to drain.


async def stop(shutdown):
    while True:
        await asyncio.sleep(0.1)  # Check 10x a second
        if shutdown.is_set():
            raise StopAsyncIteration()


def aiomain(source, track, outfd, procs, shutdown, outhwm):
    context = zmq.asyncio.Context()
    drain = context.socket(zmq.PUSH)
    drain.setsockopt(zmq.SNDHWM, outhwm)
    drain.bind(outfd)
    args = [aioreq(context, source, track, drain) for _ in range(procs)]
    args.append(stop(shutdown))
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(asyncio.gather(*args))
    except Exception:
        loop.stop()
        context.destroy(linger=0)
        loop.close()


class Requester:
    outhwm = 10

    def __init__(self, source, seed="", track="none", procs=3):
        """Create a forked asyncio frame requester object.

        Args:
            source (str): Descriptor of stream endpoint.
            seed (str, optional): File descriptor seed (to prevent ipc collisions). Defaults to "".
            track (str, optional): Video stream track name. Defaults to "none".
            procs (int, optional): Number of requester threads. Defaults to 3.
        """
        self.outfd = "ipc:///tmp/decin" + seed
        self.source, self.procs, self.track = source, procs, track
        self.shutdown = mp.Event()
        self.psargs = (
            self.source,
            self.track,
            self.outfd,
            self.procs,
            self.shutdown,
            self.outhwm,
        )
        self.ps = None

    def start(self):
        """Create and start asyncio requester threads.

        Raises:
            RuntimeError: Raised when method is called while a Requester is running.
        """
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning Requester obj")
        self.ps = mp.Process(target=aiomain, args=self.psargs)
        self.ps.daemon = True
        self.ps.start()
        print(self)

    def stop(self):
        """Join and destroy asyncio requester threads.

        Raises:
            RuntimeError: Raised when method is called while a Requester is stopped.
        """
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped Requester obj")
        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()

    def __repr__(self):
        rpr = ""
        rpr += "-----Requester-----\n"
        rpr += f"THD:\t{self.procs}\n"
        rpr += f"TRACK:\t{self.track}\n"
        rpr += f"IN:\t{self.source}\n"
        rpr += f"OUT:\t{self.outfd}\n"
        rpr += f"HWM:\t=IN> XX)::({self.outhwm} =OUT> "
        return rpr



from PIL import Image
import os

TEST_S = 0
TEST_M = 1
TEST_L = 2
IMAG_S = 3
IMAG_M = 4
IMAG_L = 5

lookup = [
    "640x480_c.png",
    "1280x720_c.png",
    "1920x1080_c.png",
    "640x480_i.png",
    "1280x720_i.png",
    "1920x1080_i.png",
]


def loadimage(enum):
    """Load a test image or a test card.

    Args:
        enum (int): One of
            pystreaming.TEST_S
            pystreaming.TEST_M
            pystreaming.TEST_L
            pystreaming.IMAG_S
            pystreaming.IMAG_M
            pystreaming.IMAG_L

    Raises:
        IndexError: Raised when received enum is not defined.

    Returns:
        PIL.Image: Image requested.
    """
    try:
        truepath = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), lookup[enum]
        )
        return Image.open(truepath)
    except IndexError:
        raise IndexError(f"Unrecognized image option: {enum}")



from pystreaming.listlib.circularlist import CircularList, Empty
from pystreaming.listlib.circulardict import CircularOrderedDict


def testcircularlist():
    try:
        cl = CircularList(0)
        assert False
    except ValueError:
        pass

    cl = CircularList(1)
    assert len(cl) == 0
    assert not cl.full()
    assert cl.__repr__() == "[None] front: 0 back: 0"

    cl.push(1)
    assert len(cl) == 1

    cl.push(2)
    assert len(cl) == 1

    cl.push(3)
    assert cl.full()
    assert len(cl) == 1
    assert cl.__repr__() == "[3] front: 0 back: 0"

    assert cl.pop() == 3
    assert not cl.full()

    cl = CircularList(5)
    assert cl.__repr__() == "[None, None, None, None, None] front: 0 back: 0"
    cl.push(0)
    cl.push(1)
    cl.push(2)
    cl.push(3)
    assert cl.__repr__() == "[0, 1, 2, 3, None] front: 0 back: 4"
    assert len(cl) == 4

    cl.push(4)
    cl.push(5)
    cl.push(6)
    assert cl.__repr__() == "[5, 6, 2, 3, 4] front: 2 back: 2"
    assert len(cl) == 5

    assert cl.pop() == 2
    assert cl.__repr__() == "[5, 6, 2, 3, 4] front: 3 back: 2"
    assert len(cl) == 4

    assert cl.pop() == 3
    assert cl.pop() == 4

    assert cl[0] == 5
    try:
        print(cl[2])
        assert False
    except IndexError:
        pass

    cl[1] = 7
    try:
        cl[2] = 8
        assert False
    except IndexError:
        pass

    assert cl.pop() == 5
    assert cl.pop() == 7
    try:
        cl.pop()
        assert False
    except Empty:
        pass


def testcirculardict():
    try:
        cd = CircularOrderedDict(0)
        assert False
    except ValueError:
        pass

    cd = CircularOrderedDict(5)
    cd.insert_end(1, 1)
    cd.insert_end(2, 2)
    cd.insert_end(3, 3)
    cd.insert_end(4, 4)
    assert len(cd) == 4
    cd.insert_end(5, 5)
    assert len(cd) == 5
    cd.insert_end(6, 6)
    assert len(cd) == 5

    for k, i in zip(cd.keys(), range(2, 7)):
        assert k == i

    cd.insert_end(3, 100)
    cd[2] = -100
    assert cd[2] == -100

    try:
        cd[7] = 7
        assert False
    except KeyError:
        pass

    assert cd.__repr__() == "OrderedDict([(2, -100), (4, 4), (5, 5), (6, 6), (3, 100)])"

    cd.delete(2)
    assert len(cd) == 4
    assert cd.pop_front() == (4, 4)
    assert len(cd) == 3
    assert cd.pop_front() == (5, 5)
    assert len(cd) == 2
    assert cd.pop_front() == (6, 6)
    assert len(cd) == 1
    assert cd.pop_front() == (3, 100)
    assert len(cd) == 0



import zmq
import time
import multiprocessing as mp
import pystreaming.video.interface as intf
from pystreaming.listlib.circularlist import CircularList, Empty
from pystreaming.video import TRACKMISS, FRAMEMISS

TIMESTEP = 0.001


def dist_ps(shutdown, infd, endpt, rcvhwm, tracks):
    context = zmq.Context()

    collector = context.socket(zmq.PULL)
    collector.setsockopt(zmq.RCVHWM, rcvhwm)
    collector.bind(infd)

    distributor = context.socket(zmq.REP)
    distributor.bind(endpt)

    if tracks is None:  # default track is "none"
        tracks = ["none"]
    queues = {track: CircularList() for track in tracks}

    while not shutdown.is_set():
        target = time.time() + TIMESTEP
        if collector.poll(0):  # returns 0 if no event, something else if there is
            buf, idx = intf.recv(collector, buf=True, flags=zmq.NOBLOCK)
            for fqueue in queues.values():
                fqueue.push((buf, idx))  # add to every buf queue

        if distributor.poll(0):  # got frame req
            track = distributor.recv().decode()
            try:
                buf, idx = queues[track].pop()
                intf.send(distributor, idx, buf=buf, flags=zmq.NOBLOCK)
            except Empty:  # Regular frame miss
                intf.send(distributor, FRAMEMISS, buf=b"nil", flags=zmq.NOBLOCK)
            except KeyError:  # Track miss
                intf.send(distributor, TRACKMISS, buf=b"nil", flags=zmq.NOBLOCK)

        missing = target - time.time()
        if missing > 0:
            time.sleep(missing)


class Distributor:
    rcvhwm = 30

    def __init__(self, endpoint, seed="", tracks=None):
        """Create a multiprocessing frame distributor object.

        Args:
            endpoint (str): Descriptor of distributor endpoint.
            seed (str, optional): File descriptor seed (to prevent ipc collisions). Defaults to "".
            tracks (list, optional): List of strings, where each string describes a track. Defaults to None.
        """
        self.infd = "ipc:///tmp/encout" + seed
        self.endpoint, self.tracks = endpoint, tracks
        self.shutdown = mp.Event()
        self.ps = None
        self.psargs = (self.shutdown, self.infd, self.endpoint, self.rcvhwm, tracks)

    def start(self):
        """Create and start a multiprocessing distributor thread.

        Raises:
            RuntimeError: Raised when method is called while a Distributor is running.
        """
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning Distributor")

        self.ps = mp.Process(target=dist_ps, args=self.psargs)
        self.ps.daemon = True
        self.ps.start()
        print(self)

    def stop(self):
        """Join and destroy the multiprocessing distributor thread.

        Raises:
            RuntimeError: Raised when method is called while a Distributor is stopped.
        """
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped Distributor")

        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()

    def __repr__(self):
        rpr = ""
        rpr += "-----Distributor-----\n"
        rpr += f"TRACK: \t{self.tracks}\n"
        rpr += f"IN: \t{self.infd}\n"
        rpr += f"OUT: \t{self.endpoint}\n"
        rpr += f"HWM: \t=IN> {self.rcvhwm})::(XX =OUT> "
        return rpr



from turbojpeg import TurboJPEG, TJSAMP_420, TJFLAG_FASTDCT
import numpy as np
import zmq
import time
import multiprocessing as mp
from functools import partial
from itertools import count
from pystreaming.video.testimages.imageloader import loadimage
import pystreaming.video.interface as intf
from pystreaming.video import QUALITY, STOPSTREAM

TIMESTEP = 0.01


def enc_ps(shutdown, infd, outfd, rcvhwm, sndhwm):
    context = zmq.Context()

    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)

    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, sndhwm)
    out.connect(outfd)

    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    encoder = partial(
        TurboJPEG().encode,
        quality=QUALITY,
        jpeg_subsample=TJSAMP_420,
        flags=TJFLAG_FASTDCT,
    )
    while not shutdown.is_set():
        target = time.time() + TIMESTEP
        if poller.poll(0):
            try:
                arr, idx = intf.recv(socket, arr=True, flags=zmq.NOBLOCK)
                intf.send(out, idx, buf=encoder(arr), flags=zmq.NOBLOCK)
            except zmq.error.Again:
                pass
        missing = target - time.time()
        if missing > 0:
            time.sleep(missing)


class Encoder:
    inhwm = 30
    rcvhwm = sndhwm = 10

    def __init__(self, context, seed="", procs=2):
        """Create a multiprocessing frame encoder object.

        Args:
            context (zmq.Context): Zmq context of calling thread.
            seed (str, optional): File descriptor seed (to prevent ipc collisions). Defaults to "".
            procs (int, optional): Number of encoding processes. Defaults to 2.
        """
        self.infd = "ipc:///tmp/encin" + seed
        self.outfd = "ipc:///tmp/encout" + seed
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.psargs = (self.shutdown, self.infd, self.outfd, self.rcvhwm, self.sndhwm)
        self.workers = []
        self.idx = 0

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.setsockopt(zmq.SNDHWM, self.inhwm)
        self.sender.bind(self.infd)

    def send(self, frame):
        """Send a frame to the encoder bank.

        Args:
            frame (np.ndarray): A frame of video. Send None to stop the stream.

        Raises:
            RuntimeError: Raised when method is called while a Encoder is stopped.
            RuntimeError: Raised if encoder processes cannot compresss frame fast enough
        """
        if self.workers == []:
            raise RuntimeError("Tried to send frame to stopped Encoder")
        try:
            if frame is None:
                intf.send(
                    self.sender, STOPSTREAM, np.zeros((10, 10, 3)), flags=zmq.NOBLOCK
                )
                ...  # Initiate STOPSTREAM
                return
            intf.send(self.sender, self.idx, arr=frame, flags=zmq.NOBLOCK)
            self.idx += 1
        except zmq.error.Again:
            raise RuntimeError("Worker processes are not processing frames fast enough")
            # change behavior to:
            # IF working, but slow, print a warning
            # If not responding, only then raise runtime error

    def start(self):
        """Create and start multiprocessing encoder threads.

        Raises:
            RuntimeError: Raised when method is called while a Encoder is running.
        """
        if self.workers != []:
            raise RuntimeError("Tried to start running Encoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=enc_ps, args=self.psargs))

        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(2)  # Allow workers to spin up
        print(self)

    def stop(self):
        """Join and destroy multiprocessing encoder threads.

        Raises:
            RuntimeError: Raised when method is called while a Encoder is stopped.
        """
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Encoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()

    def _testcard(self, cardenum, animated=False):
        """Display a testcard or a test image. Automatically starts the object
        if not already started. This method is blocking.

        Args:
            cardenum (int): One of
                pystreaming.TEST_S
                pystreaming.TEST_M
                pystreaming.TEST_L
                pystreaming.IMAG_S
                pystreaming.IMAG_M
                pystreaming.IMAG_L
            animated (bool, optional): Set True to make image rotate. Defaults to False.
        """
        testimage = loadimage(cardenum)
        if self.workers == []:
            self.start()
        if animated:
            storage = []
            for ang in range(360):
                storage.append(np.asarray(testimage.rotate(ang)))
        else:
            storage = np.asarray(testimage)
        for i in count():
            print(i)
            self.send(storage[i % 360] if animated else storage)
            time.sleep(1 / 30)

    def __repr__(self):
        rpr = ""
        rpr += "-----Encoder-----\n"
        rpr += f"PCS: \t{self.procs}\n"
        rpr += f"IN: \t{self.infd}\n"
        rpr += f"OUT: \t{self.outfd}\n"
        rpr += f"HWM: \t({self.inhwm} =IN> {self.rcvhwm})::({self.sndhwm} =OUT> "
        return rpr



import numpy as np
import zmq


def send(socket, idx, arr=None, buf=None, meta=None, flags=0):
    """Send items (idx, arr, buf, meta) through a zmq socket.

    Args:
        socket (zmq.Context.socket): Socket through which to send data.
        idx (int): Index of the sent object.
        arr (np.ndarray, optional): Numpy array to send. Defaults to None.
        buf (bytes, optional): Buffer to send. Defaults to None.
        meta (pyobj, optional): Any reasonably small picklable object. Defaults to None.
        flags (int, optional): [description]. Defaults to 0.
    """
    if arr is not None:
        md = {"dtype": str(arr.dtype), "shape": arr.shape}
        socket.send_json(md, flags=zmq.SNDMORE | flags)
        socket.send(arr, copy=False, flags=zmq.SNDMORE | flags)
    if buf is not None:
        socket.send(buf, copy=False, flags=zmq.SNDMORE | flags)
    if meta is not None:
        socket.send_pyobj(meta, flags=zmq.SNDMORE | flags)
    socket.send_pyobj(idx, flags=flags)


def recv(socket, arr=False, buf=False, meta=False, flags=0):
    """Receive data from a zmq socket.

    Args:
        socket (zmq.Context.socket): Socket through which to receive data.
        arr (bool, optional): Change to True if you expect an arr. Defaults to False.
        buf (bool, optional): Change to True if you expect a buf. Defaults to False.
        meta (bool, optional): Change to True if you expect a meta item. Defaults to False.
        flags (int, optional): Zmq flags to execute with (only zmq.NOBLOCK). Defaults to 0.

    Returns:
        list: Expected items, in the order: [arr, buf, meta, idx]
    """
    out = []
    if arr:
        md = socket.recv_json(flags=flags)
        msg = socket.recv(copy=False, flags=flags)
        arrbuf = memoryview(msg)
        out.append(np.frombuffer(arrbuf, dtype=md["dtype"]).reshape(md["shape"]))
    if buf:
        out.append(socket.recv(copy=False, flags=flags))
    if meta:
        out.append(socket.recv_pyobj(flags=flags))
    out.append(socket.recv_pyobj(flags=flags))
    return out



import zmq
import time
import pystreaming as ps


def teststartstop():
    def stopandgo(x):
        x.start()
        time.sleep(1)
        x.stop()
        time.sleep(1)
        x.start()
        time.sleep(1)
        x.stop()

    context = zmq.Context()

    devices = [
        ps.Encoder(context),
        ps.Decoder(context),
        ps.Distributor("tcp://*:5555"),
        ps.Publisher("tcp://*:5556"),
        ps.Requester("tcp://127.0.0.1:5557"),
        ps.Subscriber("tcp://127.0.0.1:5558"),
        ps.Streamer(context, "tcp://*:5559"),
        ps.Streamer(context, "tcp://*5560", mapreduce=True),
        ps.Collector(context, "tcp://*:5561"),
        ps.Collector(context, "tcp://*:5562", mapreduce=True),
        # ps.Worker(),
    ]
    for dev in devices:
        stopandgo(dev)



from turbojpeg import TurboJPEG, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import zmq
import time
from functools import partial
import multiprocessing as mp
import pystreaming.video.interface as intf

TIMESTEP = 0.01


def dec_ps(shutdown, infd, outfd, rcvmeta, sndbuf, rcvhwm, sndhwm):
    context = zmq.Context()

    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)

    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, sndhwm)
    out.connect(outfd)

    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    decoder = partial(TurboJPEG().decode, flags=(TJFLAG_FASTDCT + TJFLAG_FASTUPSAMPLE))
    while not shutdown.is_set():
        target = time.time() + TIMESTEP
        if poller.poll(0):
            try:
                package = intf.recv(socket, buf=True, meta=rcvmeta, flags=zmq.NOBLOCK)

                # Handle receiving meta
                if rcvmeta:
                    buf, meta, idx = package
                else:
                    buf, idx = package
                    meta = None

                arr = decoder(buf)

                # Handle forwarding buffer
                buf = buf if sndbuf else None

                intf.send(out, idx, arr=arr, buf=buf, meta=meta, flags=zmq.NOBLOCK)

            except zmq.error.Again:
                pass  # ignore send misses to out.

        missing = target - time.time()
        if missing > 0:
            time.sleep(missing)


class Decoder:
    outhwm = 20
    rcvhwm = sndhwm = 10

    def __init__(self, context, seed="", rcvmeta=False, sndbuf=False, procs=2):
        """Create a multiprocessing frame decoder object.

        Args:
            context (zmq.Context): Zmq context of calling thread.
            seed (str, optional): File descriptor seed (to prevent ipc collisions). Defaults to "".
            rcvmeta (bool, optional): True if we expect to receive metadata. Defaults to False.
            sndbuf (bool, optional): True if we forward the compressed frame. Defaults to False.
            procs (int, optional): Number of decoding processes. Defaults to 2.
        """
        self.infd = "ipc:///tmp/decin" + seed
        self.outfd = "ipc:///tmp/decout" + seed
        self.context, self.procs = context, procs
        self.rcvmeta, self.sndbuf = rcvmeta, sndbuf
        self.shutdown = mp.Event()
        self.psargs = (
            self.shutdown,
            self.infd,
            self.outfd,
            rcvmeta,
            sndbuf,
            self.rcvhwm,
            self.sndhwm,
        )
        self.workers = []
        self.receiver = None

    def recv(self, timeout=60_000):
        """Receive a package of data from the decoder pool.

        Args:
            timeout (int, optional): Timeout period in milliseconds.
            Set to None to wait forever. Defaults to 60_000.

        Raises:
            RuntimeError: Raised when method is called while a Decoder is stopped.
            TimeoutError: Raised when no messages are received in the timeout period.

        Returns:
            list: Either [arr, buf, meta, idx] or [arr, buf, idx] or [arr, meta, idx] or [arr, idx]
        """
        if self.workers == []:
            raise RuntimeError("Tried to receive frame from stopped Decoder")

        if self.receiver is None:
            self.receiver = self.context.socket(zmq.PULL)
            self.receiver.setsockopt(zmq.RCVHWM, self.outhwm)
            self.receiver.bind(self.outfd)

        if self.receiver.poll(timeout):
            return intf.recv(
                self.receiver,
                arr=True,
                buf=self.sndbuf,
                meta=self.rcvmeta,
                flags=zmq.NOBLOCK,
            )
        else:
            raise TimeoutError(
                f"No messages were received within the timeout period {timeout}ms"
            )

    def handler(self):
        """Yield a package of data from the decoder pool.

        Yields:
            list: Either [arr, buf, meta, idx] or [arr, buf, idx]
        """
        while True:
            yield self.recv()

    def start(self):
        """Create and start multiprocessing decoder threads.

        Raises:
            RuntimeError: Raised when method is called while a Decoder is running.
        """
        if self.workers != []:
            raise RuntimeError("Tried to start running Decoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=dec_ps, args=self.psargs))
        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(2)  # Allow workers to initialize
        print(self)

    def stop(self):
        """Join and destroy multiprocessing decoder threads.

        Raises:
            RuntimeError: Raised when method is called while a Decoder is stopped.
        """
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Decoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()

    def __repr__(self):
        rpr = ""
        rpr += "-----Decoder-----\n"
        rpr += f"PCS: \t{self.procs}\n"
        rpr += f"IN: \t{self.infd}\n"
        rpr += f"OUT: \t{self.outfd}\n"
        rpr += f"HWM: \t=IN> {self.rcvhwm})::({self.sndhwm} =OUT> {self.outhwm})"
        return rpr



from pystreaming import Encoder, Distributor, Decoder, Requester, Publisher, Subscriber
import pystreaming.video.interface as intf
import pystreaming
import zmq
import uuid
from zmq.devices import ProcessDevice


class Streamer:
    def __init__(self, context, endpoint, procs=2, mapreduce=False):
        """Video streamer with P2P and Map-Reduce functionality.

        Args:
            context (zmq.Context): Zmq context of calling thread.
            endpoint (str): Descriptor of stream endpoint.
            procs (int, optional): Number of processes devoted to encoding. Defaults to 2.
            mapreduce (bool, optional): Enable Map-Reduce streaming pattern. Defaults to False.
        """
        seed = uuid.uuid1().hex
        self.encoder = Encoder(context, seed=seed, procs=procs)
        if mapreduce:
            self.distributor = Distributor(endpoint, seed=seed)
        else:
            self.distributor = Publisher(endpoint, seed=seed)
        self.started = False

    def start(self):
        """Start internal pystreaming objects.

        Raises:
            RuntimeError: Raised when method is called while Streamer is running.
        """
        if self.started:
            raise RuntimeError("Tried to start a started Streamer")
        self.started = True
        self.encoder.start()
        self.distributor.start()

    def stop(self):
        """Cleanup and stop interal pystreaming objects.

        Raises:
            RuntimeError: Raised when method is called while the Streamer is stopped.
        """
        if not self.started:
            raise RuntimeError("Tried to stop a stopped Streamer")
        self.started = False
        self.encoder.stop()
        self.distributor.stop()

    def send(self, frame):
        """Send a video frame into the stream.

        Args:
            frame (np.ndarray): Video frame

        Raises:
            RuntimeError: Raised when method is called while the Streamer is running.
        """
        if not self.started:
            raise RuntimeError("Start the Streamer before sending frames")
        self.encoder.send(frame)

    def _testsignal(self):
        """Send a test signal through the stream. This method blocks."""
        if not self.started:
            self.start()
        self.encoder._testcard(pystreaming.TEST_M)


class Worker:
    outhwm = 30

    def __init__(self, context, source, drain, track="none", reqprocs=3, decprocs=2):
        """Map pattern in the Map-Reduce streaming pattern.

        Args:
            context (zmq.Context): Zmq context of calling thread.
            source (str): Descriptor of map-enabled stream.
            drain (str): Descriptor of Collector.
            track (str, optional): Video stream track name. Defaults to "none".
            reqprocs (int, optional): Number of request threads. Defaults to 3.
            decprocs (int, optional): Number of decoder processes. Defaults to 2.
        """
        seed = uuid.uuid1().hex
        self.requester = Requester(source, seed=seed, track=track, procs=reqprocs)
        self.decoder = Decoder(context, seed=seed, sndbuf=True, procs=decprocs)
        self.drain = context.socket(zmq.PUSH)
        self.drain.setsockopt(zmq.SNDHWM, self.outhwm)
        self.drain.connect(drain)

    def run(self, func):
        """Run map-enabled worker.

        Args:
            func (function): Stream-mapped function.
        """
        try:
            self.decoder.start()
            self.requester.start()
            for arr, buf, idx in self.decoder.handler():
                res = func(arr)
                print(res)
                try:
                    intf.send(self.drain, idx, buf=buf, meta=res, flags=zmq.NOBLOCK)
                except zmq.error.Again:
                    pass  # ignore send misses to drain.

        finally:
            self.requester.stop()
            self.decoder.stop()


class Collector:
    rcvhwm = 30
    outhwm = 30

    def __init__(self, context, endpoint, procs=2, mapreduce=False):
        """Collect frames from  video stream.

        Args:
            context (zmq.Context): Zmq context of calling thread.
            endpoint (str): Descriptor of collection endpoint.
            procs (int, optional): Number of processes devoted to decoding. Defaults to 2.
            mapreduce (bool, optional): Enable Map-Reduce streaming pattern. Defaults to False.
        """
        seed = uuid.uuid1().hex
        self.mapreduce = mapreduce
        self.startedonce = False
        if mapreduce:
            # https://het.as.utexas.edu/HET/Software/pyZMQ/api/zmq.devices.html#zmq.devices.ProcessDevice
            self.device = ProcessDevice(zmq.STREAMER, zmq.PULL, zmq.PUSH)
            self.device.setsockopt_in(zmq.RCVHWM, self.rcvhwm)
            self.device.setsockopt_out(zmq.SNDHWM, self.outhwm)
            self.device.bind_in(endpoint)
            self.device.bind_out("ipc:///tmp/decin" + seed)
        else:
            self.device = Subscriber(endpoint, seed=seed)

        self.decoder = Decoder(context, procs=procs, seed=seed, rcvmeta=mapreduce)
        self.started = False

    def start(self):
        """Start internal pystreaming objects.

        Raises:
            RuntimeError: Raised when method is called while Collector is running.
        """
        if self.started:
            raise RuntimeError("Tried to start a started Collector")
        self.started = True
        self.decoder.start()
        if self.startedonce and self.mapreduce:
            return
        self.device.start()
        self.startedonce = True

    def handler(self):
        """Returns a handler that is used for future data handling.

        Returns:
            generator: Generator that yields [arr, meta, idx] if map-reduce else [arr, idx]
        """
        if not self.started:
            self.start()
        return self.decoder.handler()

    def stop(self):
        """Cleanup and stop interal pystreaming objects.

        Raises:
            RuntimeError: Raised when method is called while the Collector is stopped.
        """
        if not self.started:
            raise RuntimeError("Tried to stop a stopped Collector")
        self.started = False
        if not self.mapreduce:
            self.device.stop()
        self.decoder.stop()



import zmq
import time
import multiprocessing as mp
import pystreaming.video.interface as intf

TIMESTEP = 0.01


def subpush_ps(shutdown, infd, outfd, rcvhwm, sndhwm):
    context = zmq.Context()

    socket = context.socket(zmq.SUB)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)
    socket.subscribe("")  # subscribe to all topics

    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, sndhwm)
    out.bind(outfd)

    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    while not shutdown.is_set():
        target = time.time() + TIMESTEP
        if poller.poll(0):
            try:
                buf, idx = intf.recv(socket, buf=True, flags=zmq.NOBLOCK)
                intf.send(out, idx, buf=buf, flags=zmq.NOBLOCK)
            except zmq.error.Again:
                pass  # Ignore misses to send out
        missing = target - time.time()
        if missing > 0:
            time.sleep(missing)  # loop takes at minimum TIMESTEP seconds


class Subscriber:
    rcvhwm = sndhwm = 10

    def __init__(self, endpoint, seed=""):
        """Create a multiprocessing subscriber.

        Connects to a zmq SUB socket and republishes through a PUSH socket.

        Args:
            endpoint (str): Descriptor of stream publishing endpoint.
            seed (str, optional): File descriptor seed (to prevent ipc collisions). Defaults to "".
        """
        self.infd = endpoint
        self.outfd = "ipc:///tmp/decin" + seed
        self.shutdown = mp.Event()
        self.psargs = (self.shutdown, self.infd, self.outfd, self.rcvhwm, self.sndhwm)
        self.ps = None

    def start(self):
        """Create and start multiprocessing subscriber threads.

        Raises:
            RuntimeError: Raised when method is called while a Subscriber is running.
        """
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning Subscriber obj")
        self.ps = mp.Process(target=subpush_ps, args=self.psargs)
        self.ps.daemon = True
        self.ps.start()
        print(self)

    def stop(self):
        """Join and destroy multiprocessing subscriber threads.

        Raises:
            RuntimeError: Raised when method is called while a Subscriber is stopped.
        """
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped Subscriber obj")
        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()

    def __repr__(self):
        rpr = ""
        rpr += "-----Subscriber-----\n"
        rpr += f"IN:\t{self.infd}\n"
        rpr += f"OUT:\t{self.outfd}\n"
        rpr += f"HWM:\t=IN> {self.rcvhwm})::({self.sndhwm} =OUT> "
        return rpr
